from typing import Dict, List, Tuple, Type, TypedDict, Union

from pydantic import BaseModel
from flask.typing import (
    HeadersValue,
    ResponseReturnValue as FlaskResponseReturnValue,
    ResponseValue as FlaskResponseValue,
    StatusCode,
)

PydanticModel = Union[Type[BaseModel], Type]

ResponseValue = Union[FlaskResponseValue, PydanticModel]

ResponseReturnValue = Union[
    FlaskResponseReturnValue,
    ResponseValue,
    Tuple[ResponseValue, HeadersValue],
    Tuple[ResponseValue, StatusCode],
    Tuple[ResponseValue, StatusCode, HeadersValue],
]


class VariableObject(TypedDict, total=False):
    enum: List[str]
    default: str
    description: str


class ServerObject(TypedDict, total=False):
    url: str
    description: str
    variables: Dict[str, VariableObject]



from .extension import FlaskSchema
from .mixins import SchemaValidationError
from .typing import ResponseReturnValue
from .validation import DataSource, validate
from .command import generate_schema_command

__all__ = (
    "DataSource",
    "FlaskSchema",
    "ResponseReturnValue",
    "SchemaValidationError",
    "validate",
    "generate_schema_command"
)



from typing import Optional

from pydantic import ValidationError


class SchemaValidationError(Exception):
    def __init__(
        self,
        validation_error: Optional[ValidationError] = None
    ) -> None:
        super().__init__()
        self.validation_error = validation_error



import json

from dataclasses import asdict, is_dataclass
from enum import Enum, auto
from functools import wraps
from typing import Any, Callable, Dict, Optional, Union, cast

from pydantic import BaseModel, ValidationError
from pydantic.dataclasses import dataclass as pydantic_dataclass, \
    is_builtin_dataclass
from pydantic.schema import model_schema
from flask import Response, jsonify, request, g
from werkzeug.datastructures import Headers
from werkzeug.exceptions import BadRequest

from .constants import SCHEMA_QUERYSTRING_ATTRIBUTE, \
    SCHEMA_REQUEST_ATTRIBUTE, SCHEMA_RESPONSE_ATTRIBUTE
from .typing import PydanticModel


class SchemaInvalidError(Exception):
    pass


class DataSource(Enum):
    FORM = auto()
    JSON = auto()


def check_query_string_schema(query_string: PydanticModel) -> PydanticModel:
    if is_builtin_dataclass(query_string):
        query_string = pydantic_dataclass(query_string).__pydantic_model__
    query_string_schema = model_schema(query_string)
    if len(query_string_schema.get("required", [])) != 0:
        raise SchemaInvalidError("Fields must be optional")
    return query_string


def check_body_schema(
    body: PydanticModel,
    source: DataSource
) -> PydanticModel:
    if is_builtin_dataclass(body):
        body = pydantic_dataclass(body).__pydantic_model__

    schema = model_schema(body)
    if source == DataSource.FORM and any(
        schema["properties"][field]["type"] == "object" for field in
            schema["properties"]
    ):
        raise SchemaInvalidError("Form must not have nested objects")
    return body


def check_response_schema(
    responses: Union[PydanticModel, Dict]
) -> Dict[int, PydanticModel]:

    if not isinstance(responses, dict):
        if is_builtin_dataclass(responses):
            responses = pydantic_dataclass(responses).__pydantic_model__
        responses = {200: responses}

    for k, v in responses.items():
        if is_builtin_dataclass(v):
            responses[k] = pydantic_dataclass(v).__pydantic_model__

    return responses


def check_response(result, response_model: Dict[int, PydanticModel]):
    status_or_headers: Union[None, int, str, Dict, list] = None
    headers: Optional[Headers] = None

    if isinstance(result, tuple):
        value, status_or_headers, headers = result + (None,) * (
            3 - len(result))
    else:
        value = result

    if isinstance(value, Response):
        value = value.get_json()

    status = 200
    if status_or_headers is not None and not isinstance(
        status_or_headers, (Headers, dict, list)
    ) and str(status_or_headers).isdigit():
        status = int(status_or_headers)

    for status_code, model_cls in response_model.items():
        if status_code != status:
            continue
        if isinstance(value, dict):
            try:
                model_value = model_cls(**value)
            except (TypeError, ValidationError) as ve:
                return jsonify(validation_error=str(ve.errors())), BadRequest
        elif type(value) == model_cls:
            model_value = value
        elif is_builtin_dataclass(value):
            model_value = model_cls(**asdict(value))
        else:
            return jsonify(validation_error="invalid response"), BadRequest
        if is_dataclass(model_value):
            return asdict(model_value), status_or_headers, headers
        else:
            model_value = cast(BaseModel, model_value)
            return model_value.dict(), status_or_headers, headers
    return result


def validate(
    query_string: Optional[PydanticModel] = None,
    body: Optional[PydanticModel] = None,
    source: DataSource = DataSource.JSON,
    validate_path: bool = False,
    responses: Union[PydanticModel, Dict[int, PydanticModel], None] = None,
    headers: Optional[PydanticModel] = None
) -> Callable:
    """
    params:
        query_string:
            the params in query
        body:
            json body or form
        source:
            the body source
        response:
            response model define
    """
    if validate_path:
        pass
    if headers is not None:
        pass

    if query_string is not None:
        query_string = check_query_string_schema(query_string)

    if body is not None:
        body = check_body_schema(body, source)

    if responses is not None:
        responses = check_response_schema(responses)

    def decorator(func: Callable) -> Callable[..., Response]:

        if query_string:
            setattr(func, SCHEMA_QUERYSTRING_ATTRIBUTE, query_string)
        if body:
            setattr(func, SCHEMA_REQUEST_ATTRIBUTE, (body, source))
        if responses:
            setattr(func, SCHEMA_RESPONSE_ATTRIBUTE, responses)

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            err = {}
            if body:
                data = request.get_json() if source == DataSource.JSON else request.form  # noqa
                try:
                    body_model = body(**data)
                except (TypeError, ValidationError) as ve:
                    err["body_params"] = str(ve)
                else:
                    g.body_params = body_model

            if query_string:
                try:
                    query_params = query_string(**request.args)
                except (TypeError, ValidationError) as ve:
                    err["query_params"] = str(ve)
                else:
                    g.query_params = query_params

            if err:
                return jsonify(validation_error=err), BadRequest.code

            result = func(*args, **kwargs)

            if responses:
                return check_response(result, responses)
            return result

        return wrapper

    return decorator



import json
from typing import Optional

import click
from flask.cli import with_appcontext
from flask import current_app as app

from .extension import _build_openapi_schema


@click.command("schema")
@click.option(
    "--output",
    "-o",
    type=click.Path(),
    help="Output the spec to a file given by a path.",
)
@with_appcontext
def generate_schema_command(output: Optional[str]) -> None:
    """
    The command which can dump json-swagger
        app.cli.add_command(generate_schema_command)
    virtualenv: flask schema
    """
    schema = _build_openapi_schema(app, app.extensions["FLASK_SCHEMA"])
    formatted_spec = json.dumps(schema, indent=2)
    if output is not None:
        with open(output, "w") as file_:
            click.echo(formatted_spec, file=file_)
    else:
        click.echo(formatted_spec)



SCHEMA_REQUEST_ATTRIBUTE = "_schema_request_schema"
SCHEMA_RESPONSE_ATTRIBUTE = "_schema_response_schemas"
SCHEMA_QUERYSTRING_ATTRIBUTE = "_schema_querystring_schema"
REF_PREFIX = "#/components/schemas/"
IGNORE_METHODS = {"OPTIONS", "HEAD"}

SWAGGER_TEMPLATE = """
<head>
  <link type="text/css" rel="stylesheet" href="{{ swagger_css_url }}">
  <title>{{ title }}</title>
</head>
<body>
  <div id="swagger-ui"></div>
  <script src="{{ swagger_js_url }}"></script>
  <script>
    const ui = SwaggerUIBundle({
      deepLinking: true,
      dom_id: "#swagger-ui",
      layout: "BaseLayout",
      presets: [
        SwaggerUIBundle.presets.apis,
        SwaggerUIBundle.SwaggerUIStandalonePreset
      ],
      showExtensions: true,
      showCommonExtensions: true,
      url: "{{ openapi_path }}"
    });
  </script>
</body>
"""

SWAGGER_JS_URL = "https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.47.1/swagger-ui-bundle.js"
SWAGGER_CSS_URL = "https://cdnjs.cloudflare.com/ajax/libs/swagger-ui/3.47.1/swagger-ui.min.css"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['schema_validator']

package_data = \
{'': ['*']}

install_requires = \
['flask>=2.0', 'pydantic>=1.8']

setup_kwargs = {
    'name': 'schema-validator',
    'version': '0.1.4',
    'description': 'A flask extension to provide schema validation with pydantic.',
    'long_description': 'flask-dantic-schema\n============\n\n#### Generate from quart-schema\n\n[![Build Status](https://app.travis-ci.com/huangxiaohen2738/flask-dantic-schema.svg?branch=main)](https://app.travis-ci.com/huangxiaohen2738/flask-dantic-schema)\n\n\n\n### Install\n\n - `pip install schema-validator`\n\n### How to use\n\n```\n    from dataclasses import dataclass\n    from datetime import datetime\n    from typing import Optional\n    from pydantic import BaseModel\n\n    from flask import Flask\n    from schema_validator import FlaskSchema, validate\n\n    app = Flask(__name__)\n    \n    FlaskSchema(app)\n    \n    OR\n    \n    schema = FlaskSchema()\n    schema.init_app(app)\n\n    @dataclass\n    class Todo:\n        task: str\n        due: Optional[datetime]\n\n    class TodoResponse(BaseModel):\n        id: int\n        name: str\n\n    @app.post("/")\n    @validate(body=Todo, responses=TodoResponse)\n    def create_todo():\n        ... # Do something with data, e.g. save to the DB\n        return dict(id=1, name="2")\n        \n    @app.put("/")\n    @validate(\n        body=Todo,\n        responses={200: TodoResponse, 400: TodoResponse}\n    )\n    def update_todo():\n        ... # Do something with data, e.g. save to the DB\n        return TodoResponse(id=1, name="123")\n       \n    app.cli.add_command(generate_schema_command)\n    \n    virtualenv:  flask schema swagger.json -> generate json swagger\n```\n',
    'author': 'hs',
    'author_email': 'huangxiaohen2738@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



import re
from collections.abc import Mapping
from dataclasses import asdict, is_dataclass
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple

from humps import camelize, decamelize
from pydantic import BaseModel
from pydantic.json import pydantic_encoder
from pydantic.schema import model_schema
from flask import Flask, current_app, render_template_string
from flask.typing import ResponseReturnValue
from flask.json import JSONDecoder, JSONEncoder

from .constants import (
    REF_PREFIX, SCHEMA_QUERYSTRING_ATTRIBUTE, SCHEMA_REQUEST_ATTRIBUTE,
    SCHEMA_RESPONSE_ATTRIBUTE, SWAGGER_CSS_URL, SWAGGER_JS_URL,
    SWAGGER_TEMPLATE, IGNORE_METHODS
)
from .typing import ServerObject
from .validation import DataSource


PATH_RE = re.compile("<(?:[^:]*:)?([^>]+)>")


class PydanticJSONEncoder(JSONEncoder):
    def default(self, object_: Any) -> Any:
        return pydantic_encoder(object_)


class CasingJSONEncoder(PydanticJSONEncoder):
    def encode(self, object_: Any) -> Any:
        if isinstance(object_, (list, Mapping)):
            object_ = camelize(object_)
        return super().encode(camelize(object_))


class CasingJSONDecoder(JSONDecoder):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, object_hook=self.object_hook, **kwargs)

    @staticmethod
    def object_hook(object_: dict) -> Any:
        return decamelize(object_)


class FlaskSchema:
    """A Flask-Schema instance.

        app = Flask(__name__)
        FlaskSchema(app)

        flask_schema = FlaskSchema()

    or

        def create_app():
            app = Flask(__name__)
            flask_schema.init_app(app)
            return app

    Arguments:
        openapi_path: The path used to serve the openapi json on, or None
            to disable documentation.
        swagger_ui_path: The path used to serve the documentation UI using
            swagger or None to disable swagger documentation.
        title: The publishable title for the app.
        version: The publishable version for the app.
    """

    def __init__(
        self,
        app: Optional[Flask] = None,
        *,
        openapi_path: Optional[str] = "/openapi.json",
        swagger_ui_path: Optional[str] = "/docs",
        title: Optional[str] = None,
        version: str = "0.1.0",
        convert_casing: bool = False,
        servers: Optional[List[ServerObject]] = None
    ) -> None:
        self.openapi_path = openapi_path
        self.swagger_ui_path = swagger_ui_path
        self.title = title
        self.version = version
        self.convert_casing = convert_casing
        self.servers = servers or []
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        app.extensions["FLASK_SCHEMA"] = self
        self.title = app.name if self.title is None else self.title
        if self.convert_casing:
            app.json_decoder = CasingJSONDecoder
            app.json_encoder = CasingJSONEncoder
        else:
            app.json_encoder = PydanticJSONEncoder

        app.config.setdefault(
            "FLASK_SCHEMA_SWAGGER_JS_URL",
            SWAGGER_JS_URL
        )
        app.config.setdefault(
            "FLASK_SCHEMA_SWAGGER_CSS_URL",
            SWAGGER_CSS_URL
        )

        if self.openapi_path is not None and app.config.get("SWAGGER_ROUTE"):
            app.add_url_rule(self.openapi_path, "openapi", self.openapi)
            if self.swagger_ui_path is not None:
                app.add_url_rule(
                    self.swagger_ui_path, "swagger_ui",
                    self.swagger_ui
                )

    def openapi(self) -> dict:
        return _build_openapi_schema(current_app, self)

    def swagger_ui(self) -> str:
        return render_template_string(
            SWAGGER_TEMPLATE,
            title=self.title,
            openapi_path=self.openapi_path,
            swagger_js_url=current_app.config["FLASK_SCHEMA_SWAGGER_JS_URL"],
            swagger_css_url=current_app.config["FLASK_SCHEMA_SWAGGER_CSS_URL"],
        )


def _split_definitions(schema: dict) -> Tuple[dict, dict]:
    new_schema = schema.copy()
    definitions = new_schema.pop("definitions", {})
    return definitions, new_schema


def _build_openapi_schema(app: Flask, extension: FlaskSchema) -> dict:
    paths: Dict[str, dict] = {}
    components = {"schemas": {}}
    for rule in app.url_map.iter_rules():
        func = app.view_functions[rule.endpoint]

        view_class = getattr(func, "view_class", None)
        if view_class is not None:
            methods = rule.methods - IGNORE_METHODS
            md = list(methods)[0].lower()
            func = getattr(view_class, md, None)

        path_object = {
            "parameters": [],
            "responses": {},
        }

        if func.__doc__ is not None:
            summary, *description = func.__doc__.splitlines()
            path_object["description"] = "\n".join(description)
            path_object["summary"] = summary

        response_models = getattr(func, SCHEMA_RESPONSE_ATTRIBUTE, {})
        for status_code, model_class in response_models.items():
            schema = model_schema(model_class, ref_prefix=REF_PREFIX)
            if extension.convert_casing:
                schema = camelize(schema)
            definitions, schema = _split_definitions(schema)
            components["schemas"].update(definitions)
            path_object["responses"][status_code] = {  # type: ignore
                "content": {
                    "application/json": {
                        "schema": schema,
                    },
                },
                "description": model_class.__doc__,
            }

        request_data = getattr(func, SCHEMA_REQUEST_ATTRIBUTE, None)
        if request_data is not None:
            schema = model_schema(request_data[0], ref_prefix=REF_PREFIX)
            if extension.convert_casing:
                schema = camelize(schema)
            definitions, schema = _split_definitions(schema)
            components["schemas"].update(definitions)

            if request_data[1] == DataSource.JSON:
                encoding = "application/json"
            else:
                encoding = "application/x-www-form-urlencoded"

            path_object["requestBody"] = {
                "content": {
                    encoding: {
                        "schema": schema,
                    },
                },
            }

        querystring_model = getattr(
            func, SCHEMA_QUERYSTRING_ATTRIBUTE, None)
        if querystring_model is not None:
            schema = model_schema(querystring_model, ref_prefix=REF_PREFIX)
            if extension.convert_casing:
                schema = camelize(schema)
            definitions, schema = _split_definitions(schema)
            components["schemas"].update(definitions)
            for name, type_ in schema["properties"].items():
                path_object["parameters"].append(  # type: ignore
                    {
                        "name": name,
                        "in": "query",
                        "schema": type_,
                    }
                )

        for name, converter in rule._converters.items():
            path_object["parameters"].append(  # type: ignore
                {
                    "name": name,
                    "in": "path",
                }
            )

        path = re.sub(PATH_RE, r"{\1}", rule.rule)
        paths.setdefault(path, {})

        for method in rule.methods:
            if method == "HEAD" or (
                method == "OPTIONS" and rule.provide_automatic_options):  # type: ignore  # noqa: E501
                continue
            paths[path][method.lower()] = path_object

    return {
        "openapi": "3.0.3",
        "info": {
            "title": extension.title,
            "version": extension.version,
        },
        "components": components,
        "paths": paths,
        "tags": [],
        "servers": extension.servers,
    }
